from scipy.stats import median_abs_deviation
import sqlite3
import numpy as np
from scipy.optimize import minimize
import pandas as pd
import logging

from ..structure.database import execute_sqlite_query, get_pandas
from ..structure.user_config import get_user_config
from ..utilities.footprint import get_combined_footprint_hash
from ..utilities.chi2_selector import get_chi2_bounds
from ..plotting.normalization_plotting import plot_normalized_star_curves


def get_fluxes(combined_footprint_hash, photometry_chi2_min, photometry_chi2_max):
    """
    Retrieves all the available star fluxes in all frames.
    If a given frame does not have a flux for this star, a NaN placeholder will be used.
    We filter by the chi2 of the fit: fluxes with an out of bounds chi2 will be replaced by a NaN
    value as well.

    :param combined_footprint_hash: int, the hash of the footprint we are working with.
    :param photometry_chi2_min: minimum acceptable chi2 value for the fit of the photometry of the star in this frame
    :param photometry_chi2_max: ditto but max
    :return: A list of frames that meet the criteria.
    """
    query = """
    SELECT s.name,
           f.id AS frame_id, 
           f.mjd,
           sff.star_gaia_id, 
           sff.combined_footprint_hash,
           IFNULL(sff.flux, 'NaN') AS flux,
           IFNULL(sff.flux_uncertainty, 'NaN') AS d_flux
    FROM 
       frames f
    JOIN star_flux_in_frame sff ON f.id = sff.frame_id 
    JOIN stars s ON sff.star_gaia_id = s.gaia_id AND sff.combined_footprint_hash = s.combined_footprint_hash
    JOIN stars_in_frames sif ON sif.star_gaia_id = s.gaia_id
                                AND sif.frame_id = f.id 
                                AND sif.combined_footprint_hash = s.combined_footprint_hash 
    WHERE 
        sff.combined_footprint_hash = ?
    AND 
        sff.chi2 BETWEEN ? AND ?
    ORDER BY 
       s.name, f.id"""
    params = (combined_footprint_hash, photometry_chi2_min, photometry_chi2_max)

    return execute_sqlite_query(query, params, is_select=True, use_pandas=True)


def update_normalization_coefficients(norm_data):
    db_path = get_user_config()['database_path']
    with sqlite3.connect(db_path, timeout=15.0) as conn:
        cursor = conn.cursor()

        # insert query with ON CONFLICT clause for bulk update
        # this handles the case where we try to insert normalization coefficients for a given frame again
        # on conflict, we update the existing record with the new coefficient values
        insert_query = """
        INSERT INTO normalization_coefficients (frame_id, combined_footprint_hash, coefficient, coefficient_uncertainty)
        VALUES (?, ?, ?, ?)
        ON CONFLICT(combined_footprint_hash, frame_id) DO UPDATE SET
        coefficient=excluded.coefficient, coefficient_uncertainty=excluded.coefficient_uncertainty
        """

        cursor.executemany(insert_query, norm_data)
        conn.commit()


def cost_function_scatter_in_frame(scaling_factors, normalized_flux_pivot, normalized_d_flux_pivot):
    """
    This is a utility for the calculation of the coef. We compare the fluxes of stars divided by their global median
    in each frame. We scale each star light-curve by a global coefficient.
    This function gives us the average scatter in each frame between the different star fluxes given a set of
    scaling factors.
    Args:
        scaling_factors: 1d array, one value per star
        normalized_flux_pivot: pivot table with frame_id as columns and star_gaia_id as rows, for fluxes
        normalized_d_flux_pivot: same but for uncertainties

    Returns: scalar, a variance representative of the scatter in each frame with the available stars in that frame.
    """
    scaled_fluxes = normalized_flux_pivot.mul(scaling_factors, axis=0)

    weights = 1 / normalized_d_flux_pivot

    weighted_means = (scaled_fluxes * weights).sum(axis=0) / weights.sum(axis=0)
    weighted_variance = (
        (weights.mul((scaled_fluxes.sub(weighted_means, axis='columns'))**2)).sum(axis=0) / weights.sum(axis=0)).sum()
    return weighted_variance


def filter_outliers(fluxes, uncertainties, threshold=3.0):
    """
    Another utility function, just filtering the very off fluxes in each frame.
    Args:
        fluxes: pandas series index by star_gaia_id
        uncertainties:  pandas series index by star_gaia_id
        threshold: float

    Returns:
        same as input types, but filtered.

    """
    median_flux = fluxes.median()
    mad = median_abs_deviation(fluxes, scale='normal')
    is_outlier = abs(fluxes - median_flux) > (threshold * mad)
    # Ensure both fluxes and uncertainties are filtered by the same outlier mask
    return fluxes[~is_outlier], uncertainties[fluxes.index[~is_outlier]]


def weighted_std(values, weights):
    """
    last utility function: weighted standard deviation for our estimation of the scatter in each coefficient.
    Args:
        values: array
        weights:  array

    Returns:
        float, some standard deviation.

    """
    isnan = np.isnan(values) + np.isnan(weights)
    values = values[~isnan]
    weights = weights[~isnan]
    average = np.average(values, weights=weights)
    variance = np.average((values-average)**2, weights=weights)
    return np.sqrt(variance)


def calculate_coefficient():
    """
    this is a routine called by the workflow manager. It interfaces with the user config and the database
    to calculate a representative norm of each frame, which will be used as a normalization coefficient later.

    Returns:
        nothing

    """
    logger = logging.getLogger('lightcurver.normalization_coefficient')
    user_config = get_user_config()

    # query initial frames, so we can calculate the footprint at hand
    # TODO for all these frames_ini requests, if by the end of building this pipeline I do not
    # TODO see a variation in the pattern, just put them in the get_combined_footprint_hash function
    # TODO so it is called only when necessary.
    frames_ini = get_pandas(columns=['id', 'image_relpath', 'exptime', 'mjd', 'seeing_pixels', 'pixel_scale'],
                            conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames_ini['id'].to_list())

    fluxes_fit_chi2_min, fluxes_fit_chi2_max = get_chi2_bounds(psf_or_fluxes='fluxes')
    df = get_fluxes(combined_footprint_hash=combined_footprint_hash,
                    photometry_chi2_min=fluxes_fit_chi2_min,
                    photometry_chi2_max=fluxes_fit_chi2_max)
    logger.info(f"Calculating a normalization coefficient using {len(df)} flux measurements.")
    # 1. normalize by median in each star -- get a 'norm' of each frame for each individual star.
    median_flux_per_star = df.groupby('star_gaia_id')['flux'].median().rename('median_flux')
    df2 = df.merge(median_flux_per_star, on='star_gaia_id')
    df2['normalized_flux'] = df2['flux'] / df2['median_flux']
    df2['normalized_d_flux'] = df2['d_flux'] / df2['median_flux']

    # we'll pivot the table to have frame_id as columns and star_gaia_id as rows
    # (useful to conduct operations within each frame)
    normalized_flux_pivot = df2.pivot(index='star_gaia_id', columns='frame_id', values='normalized_flux')
    normalized_d_flux_pivot = df2.pivot(index='star_gaia_id', columns='frame_id', values='normalized_d_flux')

    # now, these fluxes are just, for each star, 'star flux / median star flux'.
    # we could easily have scaling differences, so we'll align the curves by indivudally scaling them,
    # minimizing the dispersion in eah frame.
    # we will do this with the constraint that the mean of the coefs should be 1
    # (else, the optimizer will just set everything to 0 to minimize dispersion ...)
    constraint = ({'type': 'eq', 'fun': lambda coeffs: 1 - np.nanmean(coeffs)})
    initial_star_scaling_factors = np.ones(normalized_flux_pivot.shape[0])
    result = minimize(cost_function_scatter_in_frame, initial_star_scaling_factors,
                      args=(normalized_flux_pivot, normalized_d_flux_pivot), constraints=constraint, method='SLSQP')
    optimized_star_scaling_factors_with_reference = result.x

    # ok, scale each star light-curve by its optimized scaling
    adjusted_normalized_fluxes = normalized_flux_pivot.mul(optimized_star_scaling_factors_with_reference, axis=0)
    adjusted_normalized_d_fluxes = normalized_d_flux_pivot.mul(optimized_star_scaling_factors_with_reference, axis=0)
    logger.info(f"Fine-scaled the star light curves with coefficients:  {[round(e, 2) for e in result.x]}.")

    # now, let's eliminate the obvious outliers
    filtered_fluxes = adjusted_normalized_fluxes.copy()
    filtered_uncertainties = adjusted_normalized_d_fluxes.copy()

    # since we filtered out outliers, let us assume that the remaining uncertainties are Gaussian.
    # then, the weights that maximize the SNR in the combined quantity is the variance.
    filtered_weights = 1. / filtered_uncertainties**2
    norm_err = filtered_fluxes.columns.map(
        lambda frame_id: weighted_std(filtered_fluxes[frame_id], filtered_weights[frame_id])
    )
    norm_coeff = (filtered_fluxes.multiply(filtered_weights)).sum(axis=0) / filtered_weights.sum(axis=0)
    # restore index:
    norm_err = pd.Series(norm_err, index=filtered_fluxes.columns)
    # case with only one star: norm_err is 0, just set it to something relatively big.
    norm_err.loc[norm_err == 0.] = 0.1 * norm_coeff.loc[norm_err == 0.]

    # ok, prepare the insert into the db
    norm_data = []
    for frame_id in norm_coeff.keys():
        norm = float(norm_coeff[frame_id])
        err = float(norm_err[frame_id])
        norm_data.append((frame_id, combined_footprint_hash, norm, err))

    update_normalization_coefficients(norm_data)

    # ok, query it again in the plot function.
    plot_norm_dir = user_config['plots_dir'] / 'normalization' / str(combined_footprint_hash)
    plot_norm_dir.mkdir(exist_ok=True, parents=True)

    plot_file = plot_norm_dir / "normalization_fluxes_plot.pdf"
    plot_normalized_star_curves(combined_footprint_hash=combined_footprint_hash, save_path=plot_file)
    logger.info(f"Wrote a diagnostic plot at  {plot_file}.")





import matplotlib.pyplot as plt

from .image_plotting import plot_image
from .footprint_plotting import plot_footprints


def plot_sources(sources, image, wcs=None, save_path=None, sources_label=None,
                 kwargs_imshow=None, **kwargs_plot):
    """
    Plot the image with detected sources marked (for debugging).

    Parameters:
    sources (astropy.table.Table): Table of detected sources.
    image (numpy.ndarray): Image data, 2D array.
    wcs (astropy.wcs.WCS object): the WCS corresponding to the data. default None.
    save_path (pathlib.Path or str): path to potential save location for image.
    """
    kwargs_imshow = {} if kwargs_imshow is None else kwargs_imshow
    fig, ax = plot_image(image=image,
                         wcs=wcs,
                         save_path=None,
                         **kwargs_imshow)

    base_plot_options = {'marker': 'o',
                         'ls': 'None',
                         'mfc': 'None',
                         'color': 'red',
                         'ms': 10,
                         'alpha': 0.7}
    base_plot_options.update(kwargs_plot)

    if wcs is not None:
        ra, dec = wcs.all_pix2world(sources['xcentroid'], sources['ycentroid'], 0)
        ax.plot(ra, dec, color='red', label=sources_label,
                transform=ax.get_transform('world'),
                **base_plot_options)
    else:
        ax.plot(sources['xcentroid'], sources['ycentroid'],
                label=sources_label,
                **base_plot_options)
    if save_path is not None:
        plt.tight_layout()
        plt.savefig(save_path, bbox_inches='tight', pad_inches=0.)
        plt.close()
    else:
        return fig, ax


def plot_coordinates_and_sources_on_image(data, sources, gaia_coords, wcs, save_path, **kwargs_imshow):
    """
    This is similar to the above, but we want to focus on the quality of the WCS.
    Args:
        data: image
        sources: astropy Table
        gaia_coords: Skycoord, usually from gaia
        wcs: wcs object astropy
        save_path: where to save
        **kwargs_imshow:

    Returns:

    """

    kwargs_imshow = {} if kwargs_imshow is None else kwargs_imshow
    fig, ax = plot_image(image=data,
                         wcs=wcs,
                         save_path=None,
                         **kwargs_imshow)

    if gaia_coords is not None:
        ax.scatter(gaia_coords.ra, gaia_coords.dec, transform=ax.get_transform('world'), s=10, edgecolor='r',
                   facecolor='none', label='Gaia Stars')
    if sources is not None:
        ax.scatter(sources['x'], sources['y'], s=10, color='blue', label='Detections', alpha=0.7)

    ax.set_xlabel('RA')
    ax.set_ylabel('Dec')

    if save_path is not None:
        plt.savefig(save_path, bbox_inches='tight', pad_inches=0)
        plt.close()
    else:
        return fig, ax


def plot_footprints_with_stars(footprint_arrays, stars, save_path=None):
    """

    Args:
        footprint_arrays:  list of arrays where each array represents a footprint's corners.
        stars: pandas dataframe of stars, with columns 'name', 'ra', 'dec'
        save_path: str or path

    Returns:

    """
    fig, ax = plot_footprints(footprint_arrays, common_footprint=None, largest_footprint=None, save_path=None)
    for _, star in stars.iterrows():
        ax.plot(star['ra'], star['dec'], 'o', color='red', markersize=5, mfc='None')
        ax.text(star['ra'], star['dec'], star['name'], fontsize=8, ha='right')

    if save_path is not None:
        plt.savefig(save_path, bbox_inches='tight', pad_inches=0, dpi=300)
        plt.close()
    else:
        return fig, ax



import sqlite3
import numpy as np

from ..structure.database import execute_sqlite_query, get_pandas
from ..structure.user_config import get_user_config
from ..utilities.footprint import get_combined_footprint_hash


def r_sdss_from_gaia(gaia_like_pandas_table):
    """
        Returns r-sdss magnitudes from gaia magnitudes, from table 5.6 of
        https://gea.esac.esa.int/archive/documentation/GEDR3/Data_processing/chap_cu5pho/cu5pho_sec_photSystem/cu5pho_ssec_photRelations.html
        (see row g - r, here we do r - g, and add g.)
    Args:
        gaia_like_pandas_table: a pandas data frame with columns phot_g_mean_mag, phot_rp_mean_mag, phot_bp_mean_mag

    Returns:
        a pandas series containing r-sdss magnitudes.
    """

    bp_rp = gaia_like_pandas_table['phot_bp_mean_mag'] - gaia_like_pandas_table['phot_rp_mean_mag']
    g = gaia_like_pandas_table['phot_g_mean_mag']

    coef0 = 0.09837
    coef1 = -0.08592
    coef2 = -0.1907
    coef3 = 0.1701
    coef4 = -0.02263

    r_sdss = g + coef0 + coef1 * bp_rp + coef2 * bp_rp ** 2 + coef3 * bp_rp ** 3 + coef4 * bp_rp ** 4
    return r_sdss


def calculate_zeropoints():
    """
    for a given the current combined_footprint_hash, queries all the available star fluxes.
    compares them with gaia magnitudes to calculate a magnitude in your band
    (ONLY IMPLEMENTED FOR R-BAND SO FAR)
    TODO
    Then calculates a zeropoint for each frame, and updates the database.
    Args:

    Returns:

    """
    flux_query = """
    SELECT 
         sff.frame_id, 
         sff.flux, 
         s.gaia_id, 
         s.gmag as phot_g_mean_mag, 
         s.bmag as phot_bp_mean_mag,
         s.rmag as phot_rp_mean_mag
    FROM 
         star_flux_in_frame sff
    JOIN 
         stars s ON sff.star_gaia_id = s.gaia_id
    AND
         s.combined_footprint_hash = sff.combined_footprint_hash
    JOIN 
         frames f ON f.id = sff.frame_id
    WHERE 
         sff.combined_footprint_hash = ?
    """
    user_config = get_user_config()

    frames_ini = get_pandas(columns=['id'],
                            conditions=['plate_solved = 1', 'eliminated = 0', 'roi_in_footprint = 1'])
    combined_footprint_hash = get_combined_footprint_hash(user_config, frames_ini['id'].to_list())

    flux_data = execute_sqlite_query(flux_query, (combined_footprint_hash,), is_select=True, use_pandas=True)
    if flux_data.empty:
        return

    # so, Gaia magnitudes to r-sdss (TODO make more general, we do have band information in the database)
    flux_data['r_SDSS'] = r_sdss_from_gaia(flux_data)

    flux_data['instrumental_mag'] = -2.5 * np.log10(flux_data['flux'])
    flux_data['mag_difference'] = flux_data['r_SDSS'] - flux_data['instrumental_mag']

    # zeropoint and its uncertainty (std) for each frame
    zeropoint_results = flux_data.groupby('frame_id')['mag_difference'].agg(['median', 'std']).reset_index()
    zeropoint_results.rename(columns={'median': 'zeropoint', 'std': 'zeropoint_uncertainty'}, inplace=True)

    # and update database
    insert_query = """
    INSERT INTO approximate_zeropoints (frame_id, combined_footprint_hash, zeropoint, zeropoint_uncertainty)
    VALUES (?, ?, ?, ?)
    ON CONFLICT(frame_id, combined_footprint_hash) DO UPDATE SET
    zeropoint = excluded.zeropoint,
    zeropoint_uncertainty = excluded.zeropoint_uncertainty;
    """
    data_to_insert = [(row['frame_id'], combined_footprint_hash, row['zeropoint'], row['zeropoint_uncertainty']) for
                      _, row in zeropoint_results.iterrows()]

    db_path = get_user_config()['database_path']
    with sqlite3.connect(db_path, timeout=15.0) as conn:
        conn.executemany(insert_query, data_to_insert)




import os
from astropy.io import fits
from astropy.table import Table
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from widefield_plate_solver import plate_solve
from widefield_plate_solver.exceptions import CouldNotSolveError
import logging

from ..structure.database import execute_sqlite_query
from ..utilities.footprint import database_insert_single_footprint, get_angle_wcs


def solve_one_image(image_path, sources_path, user_config):

    sources = Table(fits.getdata(sources_path))

    if user_config['astrometry_net_api_key'] is None:
        use_api = False
    else:
        use_api = True
        os.environ['astrometry_net_api_key'] = user_config['astrometry_net_api_key']

    ra, dec = user_config['ROI_ra_deg'], user_config['ROI_dec_deg']
    plate_scale_min, plate_scale_max = user_config['plate_scale_interval']

    wcs = plate_solve(fits_file_path=image_path, sources=sources,
                      use_existing_wcs_as_guess=False,
                      use_api=use_api,
                      redo_if_done=True,  # we check for this upstream in this package
                      ra_approx=ra, dec_approx=dec,
                      scale_min=plate_scale_min, scale_max=plate_scale_max,
                      do_debug_plot=False,
                      odds_to_solve=1e8)

    return WCS(wcs)


def post_plate_solve_steps(frame_path, user_config, frame_id):
    """
    This is called after an astrometric solution has been found for an image (it is also called if the image
    is already plate solved, runs then on the existing solution)

    it
    - calculates a footprint for the image (represented by a polygon, inserted as json in the database)
    - checks if the ROI is contained by the image (if no, eliminates it)
    - has a bit of a check on the anisotropy of the pixels (bad solutions are not likely to have square pixels)
    - updates database with rotation of the field, pixel scale ...

    Args:
        frame_path: path to the fits file with WCS
        user_config: dictionary containing the user config
        frame_id: integer, database, frames column id.

    Returns:

    """
    logger = logging.getLogger("lightcurver.plate_solving")
    logger.info(f'Post plate solve steps for frame {frame_id} (path {frame_path})')
    # our object might be out of the footprint of the image!
    final_header = fits.getheader(frame_path)
    # replace the wcs above with the WCS we saved in the header of the image (contains naxis)
    wcs = WCS(final_header)
    in_footprint = user_config['ROI_SkyCoord'].contained_by(wcs)
    if in_footprint:
        execute_sqlite_query(query="UPDATE frames SET roi_in_footprint = ? WHERE id = ?",
                             params=(1, frame_id), is_select=False)
    # also, let us save the actual footprint
    footprint_array = wcs.calc_footprint()
    database_insert_single_footprint(frame_id, footprint_array)
    # and let us compute the pixel scale!
    psx, psy = proj_plane_pixel_scales(wcs)
    # these are most likely in deg/pixel. astropy says that wcs should carry a cunit attribute,
    # but it does not. Anyway, let us assume deg/pixel -- never seen anything else when working with wcs
    # of wide field images.
    anisotropy = float(abs(psx - psy) / (psx + psy))
    suspicious_astrometry = abs(psx - psy) / (psx + psy) > float(user_config['max_pixel_anisotropy'])
    if suspicious_astrometry:
        message = "Your pixels are more rectangular than your config tolerance! Flagging (eliminating) this frame."
        message += f"Anisotropy: {anisotropy:.01%}, path: {frame_path}, db id: {frame_id})."
        logger.info(message)
        execute_sqlite_query(query="""UPDATE 
                                          frames 
                                      SET 
                                          eliminated = 1,
                                          comment='suspicious_plate_solved'
                                      WHERE 
                                          id = ?""",
                             params=(frame_id,), is_select=False)
    angle_to_north = get_angle_wcs(wcs)
    pixel_scale = 0.5 * (psx + psy) * 3600  # to arcsecond / pixel
    # first, set the pixel scale
    execute_sqlite_query(query="UPDATE frames SET pixel_scale = ? WHERE id = ?",
                         params=(pixel_scale, frame_id), is_select=False)
    # then, use it to compute the seeing in arcseconds, and insert the angle at the same time to combine queries
    execute_sqlite_query(query="""UPDATE    
                                      frames 
                                  SET 
                                      seeing_arcseconds = pixel_scale * seeing_pixels,
                                      angle_to_north = ?
                                  WHERE 
                                      id = ?""",
                         params=(angle_to_north, frame_id), is_select=False)
    logger.info(f'Updated pixel scale: {pixel_scale:.03f}"/pixel for frame {frame_id} (path {frame_path}).')


def solve_one_image_and_update_database(image_path, sources_path, user_config, frame_id):
    """
    solves image using the sources in sources_path, then adds useful things to the database.
    If already solved according to user_config, just does the database part.
    Args:
        image_path: path to fits file containing the image
        sources_path: path to fits file containing the sources extracted from the image
        user_config: dictionary read by the pipeline
        frame_id: the database frame id
    Returns:
        nothing
    """
    logger = logging.getLogger("lightcurver.plate_solving")
    if not user_config['already_plate_solved']:
        logger.info(f'Attempting astrometric solution for frame {frame_id} (path: {image_path}).')
        try:
            wcs = solve_one_image(image_path, sources_path, user_config)
            success = wcs.is_celestial
        except CouldNotSolveError:
            success = False
    else:
        logger.info(f'Frame {frame_id} (path: {image_path}) is already solved according to user config.')
        success = True

    if success:
        post_plate_solve_steps(frame_id=frame_id, frame_path=image_path, user_config=user_config)
    # at the end, set the image to plate solved in db, and flag it as having had a plate solve attempt.
    execute_sqlite_query(query="UPDATE frames SET plate_solved = ?, attempted_plate_solve = 1 WHERE id = ?",
                         params=(1 if success else 0, frame_id), is_select=False)



import yaml
from importlib import resources
from collections import deque
import logging
from datetime import datetime


from ..structure.user_config import get_user_config
from ..structure.database import initialize_database
from ..processes.cutout_making import extract_all_stamps
from ..processes.star_querying import query_gaia_stars
from ..processes.psf_modelling import model_all_psfs
from ..processes.star_photometry import do_star_photometry
from ..processes.normalization_calculation import calculate_coefficient
from ..processes.roi_deconv_file_preparation import prepare_roi_deconv_file
from ..processes.roi_modelling import do_deconvolution_of_roi
from ..processes.alternate_plate_solving_with_gaia import alternate_plate_solve
from ..utilities.zeropoint_from_gaia import calculate_zeropoints
from ..structure.exceptions import TaskWasNotSuccessful
from .task_wrappers import (read_convert_skysub_character_catalog,
                            plate_solve_all_frames, calc_common_and_total_footprint_and_save)
from .state_checkers import check_plate_solving


def setup_base_logger():
    time_now = datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
    user_config = get_user_config()
    log_dir = user_config['workdir'] / 'logs'
    log_file_path = str(log_dir / f"{time_now}.log")

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    base_logger = logging.getLogger("lightcurver")
    file_handler = logging.FileHandler(log_file_path)
    file_handler.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    base_logger.addHandler(file_handler)
    base_logger.setLevel(logging.INFO)


class WorkflowManager:
    """
    A small class that will run our tasks. It serves the purpose of running the tasks
    in the right order given the dependencies in the pipeline_dependency_graph.yaml file.
    We can also use it to assign different versions of the tasks based on the user config.

    todo: implement post checks for each tasks
    todo: implement pre checks so we can skip tasks from the workflow manager? for now pre-checks are done within each
          task.
    """
    def __init__(self, logger=None):
        self.user_config = get_user_config()
        with resources.open_text('lightcurver.pipeline', 'pipeline_dependency_graph.yaml') as file:
            self.pipe_config = yaml.safe_load(file)
        self.task_graph = {}
        self.build_dependency_graph()
        # some tasks can be done in multiple ways, let's define this here
        if self.user_config['plate_solving_strategy'] == 'plate_solve':
            plate_solve_function = plate_solve_all_frames
        elif self.user_config['plate_solving_strategy'] == 'alternate_gaia_solve':
            plate_solve_function = alternate_plate_solve
        else:
            raise AssertionError("The config's plate_solving_strategy should be plate_solve or alternate_gaia_solve")
        self.task_attribution = {
            'initialize_database': initialize_database,
            'read_convert_skysub_character_catalog': read_convert_skysub_character_catalog,
            'plate_solving': plate_solve_function,
            'calculate_common_and_total_footprint': calc_common_and_total_footprint_and_save,
            'query_gaia_for_stars': query_gaia_stars,
            'stamp_extraction': extract_all_stamps,
            'psf_modeling': model_all_psfs,
            'star_photometry': do_star_photometry,
            'calculate_normalization_coefficient': calculate_coefficient,
            'calculate_absolute_zeropoints': calculate_zeropoints,
            'prepare_calibrated_cutouts': prepare_roi_deconv_file,
            'model_calibrated_cutouts': do_deconvolution_of_roi,
        }

        self.post_task_attribution = {
            'plate_solving': check_plate_solving
        }
        assert set(self.task_attribution.keys()) == set([entry['name'] for entry in self.pipe_config['tasks']])

        if logger is None:
            logger = logging.getLogger(__name__)
            setup_base_logger()
        self.logger = logger

    def build_dependency_graph(self):
        for task in self.pipe_config['tasks']:
            task_name = task['name']
            self.task_graph[task_name] = {'dependencies': set(task['dependencies']), 'next': []}
            for dep in task['dependencies']:
                if dep in self.task_graph:
                    self.task_graph[dep]['next'].append(task_name)
                else:
                    self.task_graph[dep] = {'dependencies': set(), 'next': [task_name]}

    def topological_sort(self):
        """
            Getting the tasks in the right order, just in case we have multiple dependencies in the future.

            Returns: list of tasks in the right order.
        """
        #
        in_degree = {task: 0 for task in self.task_graph}
        for task in self.task_graph:
            for next_task in self.task_graph[task]['next']:
                in_degree[next_task] += 1

        queue = deque([task for task in in_degree if in_degree[task] == 0])
        sorted_tasks = []

        while queue:
            task = queue.popleft()
            sorted_tasks.append(task)
            for next_task in self.task_graph[task]['next']:
                in_degree[next_task] -= 1
                if in_degree[next_task] == 0:
                    queue.append(next_task)

        if len(sorted_tasks) != len(self.task_graph):
            raise Exception("A cycle was detected in the task dependencies, or a task is missing.")

        return sorted_tasks

    def run(self):
        self.logger.info(f"Workflow manager: will run all tasks. Working directory:  {self.user_config['workdir']}")
        sorted_tasks = self.topological_sort()
        for task_name in sorted_tasks:
            task = next((item for item in self.pipe_config['tasks'] if item['name'] == task_name), None)
            if task:
                self.execute_task(task)
            post_check = self.post_task_attribution.get(task_name, None)
            if post_check:
                success, message = post_check()
                if not success:
                    self.logger.error(
                        f'Post-check failed for {task_name}. Stopping pipeline with message: {message}'
                    )
                    raise TaskWasNotSuccessful(message)
                else:
                    self.logger.info(f'Post-check successful for task {task_name}, with message: {message}')

    def execute_task(self, task):
        # Assume task_func is a callable for simplicity
        task_func = self.task_attribution.get(task['name'])
        self.logger.info(
            f"Running task {task['name']}. Working directory:  {self.user_config['workdir']}"
        )
        task_func()

    def get_tasks(self):
        return sorted(list(self.task_attribution.keys()))






# This file wraps around the processes defined in the processes subpackage.
# The wrappers typically determine which frames / regions / psfs (depending on task) need processing
# before proceeding, and adds structure around multiprocessing when needed.
# This is not needed for all processes.
from multiprocessing import Pool, Manager
import os
import numpy as np
import pandas as pd
import logging
import logging.handlers
from pathlib import Path
import functools
import json


from ..structure.user_config import get_user_config
from ..structure.database import get_pandas, execute_sqlite_query
from ..processes.frame_importation import process_new_frame
from ..processes.plate_solving import solve_one_image_and_update_database
from ..utilities.footprint import (calc_common_and_total_footprint, get_frames_hash,
                                   save_combined_footprints_to_db, identify_and_eliminate_bad_pointings)
from ..plotting.footprint_plotting import plot_footprints
from ..processes.star_extraction import extract_sources_from_sky_sub_image


def worker_init(log_queue):
    logger = logging.getLogger(f"worker-{os.getpid()}")
    logger.setLevel(logging.INFO)
    q_handler = logging.handlers.QueueHandler(log_queue)
    logger.addHandler(q_handler)
    logger.propagate = False


def log_process(func):
    @functools.wraps(func)
    def wrapper(args):
        frame_id_for_logger = args[-1]
        logger = logging.getLogger(f"Process-{os.getpid()}")
        logger.info(f"{func.__name__} .... Processing image with ID {frame_id_for_logger}")
        return func(*args[:-1])  # execute original function without the last arg (used for logging)
    return wrapper


@log_process
def process_new_frame_wrapper(*args):
    process_new_frame(*args)


def read_convert_skysub_character_catalog():
    # boiler plate logger setup
    log_queue = Manager().Queue()
    base_logger = logging.getLogger("lightcurver")
    listener = logging.handlers.QueueListener(log_queue, *base_logger.handlers)
    listener.start()
    logger = logging.getLogger("lightcurver.importation")

    # find the new frames, we compare on file name!
    user_config = get_user_config()
    available_frames = sum([list(raw_dir.glob('*')) for raw_dir in user_config['raw_dirs']], start=[])
    df_available_frames = pd.DataFrame({'frame_name': [frame.name for frame in available_frames]})
    already_imported = get_pandas(columns=['original_image_path', 'id'])
    if not already_imported.empty:
        already_imported['name'] = already_imported.apply(lambda row: Path(row['original_image_path']).name, axis=1)
    else:
        # just a decoy
        already_imported['name'] = pd.Series(dtype='str')
    new_frames_df = df_available_frames[~df_available_frames['frame_name'].isin(already_imported['name'])]
    new_frames = [frame for frame in available_frames if frame.name in new_frames_df['frame_name'].tolist()]
    logger.info(f"Importing {len(new_frames)} new frames from directories {user_config['raw_dirs']}.")
    logger.info(f"Will write them to {user_config['workdir'] / 'frames'}")
    logger.info(f"Database will be at {user_config['workdir'] / 'database.sqlite3'}")

    with Pool(processes=user_config['multiprocessing_cpu_count'],
              initializer=worker_init, initargs=(log_queue,)) as pool:
        pool.map(process_new_frame_wrapper, [
            (new_frame,
             user_config,
             new_frame)  # duplicating so to have an identifier for logger.
            for new_frame in new_frames
        ])

    listener.stop()


@log_process
def solve_one_image_and_update_database_wrapper(*args):
    solve_one_image_and_update_database(*args)


def plate_solve_all_frames():
    # boiler plate logger setup
    log_queue = Manager().Queue()
    base_logger = logging.getLogger("lightcurver")
    listener = logging.handlers.QueueListener(log_queue, *base_logger.handlers)
    listener.start()
    logger = logging.getLogger("lightcurver.plate_solving")

    user_config = get_user_config()
    workdir = Path(user_config['workdir'])

    # so, we select our frames to plate solve depending on the user config.
    if user_config['plate_solve_frames'] == 'all_not_eliminated':
        conditions = ['eliminated = 0']
        logger.info(f"Processing all the frames (even the ones already solved) that are not flagged as eliminated.")
    elif user_config['plate_solve_frames'] == 'all_never_attempted':
        conditions = ['eliminated = 0', 'attempted_plate_solve = 0']
        logger.info(f"Processing all the frames that do not have a solve attempt yet.")
    elif user_config['plate_solve_frames'] == 'all_not_plate_solved':
        conditions = ['eliminated = 0', 'plate_solved = 0']
        logger.info(f"Processing all the frames that are not plate solved, even those that were already attempted.")
    else:
        raise ValueError(f"Not an expected selection strategy for frames to solve: {user_config['plate_solve_frames']}")

    frames_to_process = get_pandas(columns=['id', 'image_relpath', 'sources_relpath'],
                                   conditions=conditions)
    logger.info(f"Ready to plate solve {len(frames_to_process)} frames.")

    with Pool(processes=user_config['multiprocessing_cpu_count'],
              initializer=worker_init, initargs=(log_queue,)) as pool:
        pool.map(solve_one_image_and_update_database_wrapper, [
            (workdir / row['image_relpath'],
             workdir / row['sources_relpath'],
             user_config,
             row['id'],
             row['id'])  # duplicating row['id'] for logger naming
            for index, row in frames_to_process.iterrows()
        ])

    listener.stop()


def calc_common_and_total_footprint_and_save():
    """
    verifies whether the footprint was already calculated for the set of frames at hand
    if no, calculates it and stores it.

    Returns: None

    """
    logger = logging.getLogger("lightcurver.combined_footprint_calculation")
    # so, before we do anything, let us eliminate the really obvious bad pointings.
    identify_and_eliminate_bad_pointings()
    # ok, keep going.
    query = """
    SELECT frames.id, footprints.polygon
    FROM footprints 
    JOIN frames ON footprints.frame_id = frames.id 
    WHERE frames.eliminated != 1;
    """
    results = execute_sqlite_query(query)
    frames_ids = [result[0] for result in results]
    frames_hash = get_frames_hash(frames_ids)
    # check if already exists
    count = execute_sqlite_query("SELECT COUNT(*) FROM combined_footprint WHERE hash = ?",
                                 params=(frames_hash,))[0][0]
    if count > 0:
        logger.info(f'This combined footprint (hash {frames_hash}) was already calculated.')
        return
    logger.info(f'Calculating combined footprint (hash {frames_hash}) (loaded polygons from database, now combining).')
    polygon_list = [np.array(json.loads(result[1])) for result in results]
    common_footprint, largest_footprint = calc_common_and_total_footprint(polygon_list)

    user_config = get_user_config()
    plots_dir = user_config['plots_dir']
    footprints_plot_path = plots_dir / 'footprints.jpg'
    plot_footprints(polygon_list, common_footprint, largest_footprint, save_path=footprints_plot_path)
    logger.info(f'Combined footprint plot (hash {frames_hash}) saved at {footprints_plot_path}.')

    # ok, save it
    save_combined_footprints_to_db(frames_hash, common_footprint, largest_footprint)
    logger.info(f'Combined footprint with (hash {frames_hash} saved to db')


@log_process
def extract_sources_from_sky_sub_image_wrapper(*args):
    extract_sources_from_sky_sub_image(*args)


def source_extract_all_images(conditions=None):
    """
    This is not called directly in the pipeline, but can be useful if you want to re-extract all the sources
    with different parameters.
    So, unlike a routine called by the pipeline, this one takes an argument.

    :param conditions: list of strings, e.g. ['eliminated = 0', 'plate_solved = 0']. Default: None.
                       To filter what frames will be source-extracted again.
    :returns: Nothing

    """
    log_queue = Manager().Queue()
    base_logger = logging.getLogger("lightcurver")
    listener = logging.handlers.QueueListener(log_queue, *base_logger.handlers)
    listener.start()
    logger = logging.getLogger("lightcurver.source_extraction")

    user_config = get_user_config()
    workdir = Path(user_config['workdir'])
    frames_to_process = get_pandas(columns=['id', 'image_relpath', 'sources_relpath',
                                            'exptime', 'background_rms_electron_per_second'],
                                   conditions=conditions)
    logger.info(f"Extracting sources from {len(frames_to_process)} frames.")

    with Pool(processes=user_config['multiprocessing_cpu_count'],
              initializer=worker_init, initargs=(log_queue,)) as pool:
        pool.map(extract_sources_from_sky_sub_image_wrapper, [
            (workdir / row['image_relpath'],
             workdir / row['sources_relpath'],
             user_config['source_extraction_threshold'],
             user_config['source_extraction_min_area'],
             row['exptime'],
             row['background_rms_electron_per_second'],
             user_config['plots_dir'] / 'source_extraction' / f"{Path(row['image_relpath']).stem}.jpg",
             row['id'])  # for logger naming
            for index, row in frames_to_process.iterrows()
        ])

    listener.stop()



import numpy as np
from shapely.geometry import Polygon, mapping
from functools import reduce
import json
from astropy.io import fits
from astropy.wcs import WCS

from ..structure.database import execute_sqlite_query, get_pandas
from ..structure.user_config import get_user_config


def get_combined_footprint_hash(user_config, frames_id_list):
    """
    Calculates the hash of the combined footprint of the frames whose id is in frames_id_list.
    Args:
        user_config: dictionary obtained from structure.config.get_user_config
        frames_id_list: list of integers, which frames are we using.

    Returns:
        frames_hash, integer

    """
    if user_config['star_selection_strategy'] != 'ROI_disk':
        # then it depends on the frames we're considering.
        frames_hash = get_frames_hash(frames_id_list)
    else:
        # if ROI_disk, it does not depend on the frames: unique region defined by its radius.
        return hash(user_config['ROI_disk_radius_arcseconds'])


def calc_common_and_total_footprint(list_of_footprints):
    """
    Calculate the common (intersection) and largest (union) footprints from a list of numpy arrays: products
    of astropy.wcs.WCS.calc_footprint.
    Then determines both the intersection and union of these
    footprints using Shapely Polygons. The intersection represents the area common to all frames,
    while the union covers the total area spanned by any of the frames.

    Parameters:
    - list_of_fits_paths: A list of paths to FITS files.

    Returns:
    - wcs_objects: A list of WCS objects corresponding to the FITS files.
    - common_footprint: A Shapely Polygon representing the intersected footprint common to all WCS footprints.
    - largest_footprint: A Shapely Polygon representing the union of all WCS footprints.
    """
    wcs_footprints = list_of_footprints

    polygons = [Polygon(footprint) for footprint in wcs_footprints]

    common_footprint = reduce(lambda x, y: x.intersection(y), polygons)
    largest_footprint = reduce(lambda x, y: x.union(y), polygons)

    return (common_footprint.simplify(tolerance=0.001, preserve_topology=True),
            largest_footprint.simplify(tolerance=0.001, preserve_topology=True))


def database_insert_single_footprint(frame_id, footprint_array):
    polygon_list = footprint_array.tolist()
    polygon_str = json.dumps(polygon_list)

    execute_sqlite_query(query="INSERT OR REPLACE INTO footprints (frame_id, polygon) VALUES (?, ?)",
                         params=(frame_id, polygon_str),
                         is_select=False)


def database_get_footprint(frame_id):
    result = execute_sqlite_query(query="SELECT polygon FROM footprints WHERE frame_id = ?",
                                  params=(frame_id,),
                                  is_select=True)[0]

    polygon_list = json.loads(result)
    footprint_polygon = np.array(polygon_list)

    return footprint_polygon


def get_frames_hash(frames_ids):
    """
    when calculating footprints, we need a way to identify which footprint was calculated from which frames.
    I don't want to deal with the relational many-to-many situation that will arise in a relational database,
    so let's calculate a hash of the integers of the frames that were used to calculate a footprint.
    then to check for a footprint, which can just query the hash.
    Args:
        frames_ids: list of integers, frames.id in the database

    Returns:
        a text hash

    """
    assert len(set(frames_ids)) == len(frames_ids), "Non-unique frame ids passed to this function"
    sorted_frame_ids = sorted(frames_ids)
    frame_ids_tuple = tuple(sorted_frame_ids)
    return hash(frame_ids_tuple)


def save_combined_footprints_to_db(frames_hash, common_footprint, largest_footprint):

    common_str = json.dumps(mapping(common_footprint))
    largest_str = json.dumps(mapping(largest_footprint))
    save_query = "INSERT INTO combined_footprint (hash, largest, common) VALUES (?, ?, ?)"
    execute_sqlite_query(save_query,
                         params=(frames_hash, largest_str, common_str),
                         is_select=False)


def load_combined_footprint_from_db(frames_hash):

    query = "SELECT largest, common FROM combined_footprint WHERE hash = ?"
    result = execute_sqlite_query(query,
                                  params=(frames_hash,),
                                  is_select=True)[0]
    largest = json.loads(result[0])
    common = json.loads(result[1])
    if result:
        return largest, common
    else:
        return None


def check_in_footprint_for_all_images():
    """
    Just a wrapper for running the footprint check. Can be useful to manually execute in some cases.
    Will
     - load the footprints from db or load all the headers of the plate solved images depending on function argument
     - check if the ROI coord is in the footprint defined by the WCS
     - update the frames table.
    We skip the footprints table, as this is the ROI and not a simple star we want a better check
    than the simple gnonomic projection we are forced to rely on when using the footprint table.

    Returns:
        Nothing
    """
    frames_to_process = get_pandas(columns=['id', 'image_relpath'],
                                   conditions=['plate_solved = 1', 'eliminated = 0'])
    user_config = get_user_config()

    for i, frame in frames_to_process.iterrows():
        frame_id = frame['id']
        frame_path = user_config['workdir'] / frame['image_relpath']
        final_header = fits.getheader(frame_path)
        wcs = WCS(final_header)
        in_footprint = user_config['ROI_SkyCoord'].contained_by(wcs)
        execute_sqlite_query(query="UPDATE frames SET roi_in_footprint = ? WHERE id = ?",
                             params=(int(in_footprint), frame_id), is_select=False)


def identify_and_eliminate_bad_pointings():
    """
    Called after calculating the footprints. Will identify pointings that are ~really~ different, and
    flag them in the database ('eliminated = 1', 'comment = "bad_pointing"')
    Returns: nothing

    """

    select_query = """
    SELECT frames.id, footprints.polygon
    FROM footprints 
    JOIN frames ON footprints.frame_id = frames.id
    WHERE frames.eliminated != 1;
    """
    update_query = """
    UPDATE frames
    SET comment = 'bad_pointing', eliminated = 1
    WHERE id = ?;
    """

    results = execute_sqlite_query(select_query, is_select=True, use_pandas=True)
    mean_positions = []

    for i, row in results.iterrows():
        frame_id = row['id']
        polygon = row['polygon']
        polygon = np.array(json.loads(polygon))
        mean_position = np.mean(polygon, axis=0)
        mean_positions.append((frame_id, mean_position))

    all_means = np.array([pos for _, pos in mean_positions])
    overall_mean = np.mean(all_means, axis=0)

    # distance of each frame's mean position from the overall mean
    deviations = [(frame_id, np.linalg.norm(mean_pos - overall_mean)) for frame_id, mean_pos in mean_positions]

    # threshold
    deviation_values = [dev for _, dev in deviations]
    mean_deviation = np.mean(deviation_values)
    std_deviation = np.std(deviation_values)
    threshold = mean_deviation + 5 * std_deviation  # quite a generous threshold.

    # flag frames with significant deviation
    bad_frames = [frame_id for frame_id, dev in deviations if dev > threshold]

    for frame_id in bad_frames:
        execute_sqlite_query(update_query, params=(frame_id,))


def get_angle_wcs(wcs_object):
    """
     Takes a WCS object, and returns the angle in degrees to the North (so, angle relative to "North up, East left")
    Args:
        wcs_object: astropy WCS object

    Returns:
        angle: float, angle in degrees.
    """

    if hasattr(wcs_object.wcs, 'cd'):
        matrix = wcs_object.wcs.cd
    elif hasattr(wcs_object.wcs, 'pc'):
        matrix = wcs_object.wcs.pc
    else:
        raise ValueError("Neither CD nor PC matrix found in WCS.")

    cd1_1, cd1_2 = matrix[0, 0], matrix[0, 1]
    cd2_1, cd2_2 = matrix[1, 0], matrix[1, 1]

    angle = np.arctan2(-cd1_2, cd2_2) * 180.0 / np.pi

    return angle
